/*DIPhisto.c (A. Christidis)*/

#include "LoadImg.h"
#include "DIPhisto.h"
#include "ShowAux.h"

enum {R, G, B};

/*Merke: extern darf nicht initialisiert werden! (Deklaration vs. Definition)*/
extern DIPimg   img, aoi;
extern int      aoiX0, aoiY0;
extern GLdouble zoom;
extern int      mainWin, auxWin[WINADMIN];

/***************************************************************************/
  void picStats (DIPimg *pic, int *min, int *Max, int *maxNpix, int *hist)
/***************************************************************************/
/*Statistik der Pixel in einem Bild: Ermittlung des minimalen bzw. maximalen 
  Grauwertes <min> bzw. <Max>, der Anzahl des am haeufigsten auftretenden 
  Wertes <maxNpix> und des Histogramms (Haeufigkeit jedes Grauwerts) <hist>
  Wird fuer <hist> ein NULL-Pointer uebergeben, so unterbleiben 
  Wertzuweisungen an <hist>.*/
{ int   jx=0, jy=0, jcmp=0;
#ifdef  MORE_HIST
#endif//MORE_HIST

  /*Ermittlung von min, Max, maxNpix:*/
  for (jcmp=0; jcmp<3; jcmp++)
  { for(jy=0; jy<pic->Height; jy++) 
    { for(jx=0; jx<pic->Width; jx++) 
      {
#ifdef  MORE_HIST
#endif//MORE_HIST
      }
    }
   }
  return;
}

/***************************************************************************/
  void optiCont (DIPimg *picIn, DIPimg *picOut)
/***************************************************************************/
/*Bild-Kontrast anhand der Pixel-Statistik eines Bildes maximieren:
  (255-0+1=)256 Helligkeitswerte verteilen ber (Max-min+1) Helligkeitswerte.
  Werte, die im Bild nicht vorkommen, liegen danach ausserhalb [0...255]!*/
{          int  jx=0, jy=0, min=0, Max=0, maxNpix=0, lut[256], *hist=NULL;
  unsigned int  ipixel=0;
  unsigned char *pixel=(unsigned char *)&ipixel;

  /*Initialisierung:*/
  for(jx=0; jx<256; jx++) lut[jx] = 0; 

  /*Ermittlung der benoetigten Statistik:*/
   picStats (picIn, &min, &Max, &maxNpix, hist);

  /*Look-Up-Table mit der Grauwert-Transformation: 
   (G(neu)-0+1)/(G(alt)-min+1) = (255-0+1)/(Max-min+1) // frueherer Versuch
   (G(neu)-0)/(255-0) = (G(alt)-min)/(Max-min)*/

#ifdef  MORE_HIST
#endif//MORE_HIST

 /*Transformation der Helligkeitswerte:*/
  for(jy=0; jy<picOut->Height; jy++) 
  { for(jx=0; jx<picOut->Width; jx++) 
    { 
      ipixel = picIn->Data[jy*picIn->Width+jx];
#ifdef  MORE_HIST
#endif//MORE_HIST
      picOut->Data[jy*picIn->Width+jx] = ipixel;
    }
  }
  return;
}

/***************************************************************************/
  void Hist (void)
/***************************************************************************/
/*Histogramm des Bildes <pic>:*/
{ int jx=0, jy=0, jcmp=0, border=5, min=0, Max=0, maxNpix=0, hist[3*256];
  float scale=1.f;

  /*Linienstaerke:*/
  glLineWidth (1.);

  /*Bild-Statistik:*/
  picStats (&aoi, &min, &Max, &maxNpix, hist);
  /*Skalierungsfaktor <scale> bringt auch haeufigsten Wert ins Fenster: 
    mit Kantenlenge HIST_DIM:*/
#ifdef  MORE_HIST
#endif//MORE_HIST

  /*Gefahr einer Endlos-Schleife, da Hist() auch Callback:*/
  if (!auxWin[R] || !auxWin[G] || !auxWin[B])
      InitAuxWins (WINADMIN, HIST_DIM+2*border, HIST_DIM+2*border, Hist);

  /*Histogramm-Darstellung in den Hilfsfenstern:*/
  for (jcmp=0; jcmp<3; jcmp++)
  { glutSetWindow(auxWin[jcmp]);
    glutShowWindow();
    glClear(GL_COLOR_BUFFER_BIT);
    glutReshapeWindow (HIST_DIM+2*border, HIST_DIM+2*border); 
    DrawIn2D (HIST_DIM+2*border, HIST_DIM+2*border);
    if (jcmp == 0) { glutSetWindowTitle("Histogram RED"); 
                     glColor3ub(255, 0, 0); } else
    if (jcmp == 1) { glutSetWindowTitle("Histogram GREEN"); 
                     glColor3ub(0, 255, 0); } else
    if (jcmp == 2) { glutSetWindowTitle("Histogram BLUE"); 
                     glColor3ub(0, 0, 255); }

    /*Balken-Diagramm:*/
    for(jy=min; jy<=Max; jy++) 
    { glBegin (GL_LINES);
#ifdef  MORE_HIST
#endif//MORE_HIST
      glEnd();
    }
  }
  glFinish(); 

  /*Rahmen um AOI am Bild selbst:*/
  glutSetWindow(mainWin);
  glColor3ub(255, 255, 0);
  /*Fenstersystem die aktuelle Zeichenflaeche mitteilen:*/
  DrawIn2D (zoom*aoi.Width, zoom*aoi.Height);

  glBegin (GL_LINE_LOOP);
    glVertex2i((int)(zoom*aoiX0), (int)(zoom*aoiY0)); 
               glVertex2i((int)(zoom*(aoiX0+aoi.Width)-1), (int)(zoom*aoiY0)); 
    glVertex2i((int)(zoom*(aoiX0+aoi.Width)-1), (int)(zoom*(aoiY0+aoi.Height)-1)); 
               glVertex2i((int)(zoom*aoiX0), (int)(zoom*(aoiY0+aoi.Height)-1)); 
  glEnd();
  glFinish(); 

  return;
}

/***************************************************************************/
#ifdef  MORE_HIST
#else //MORE_HIST
  void InitAuxWins (int nWins, int width, int height, int (*callback)(float))
#endif//MORE_HIST
/***************************************************************************/
{ int jj=0;

  /*Schutz vor Ueberlauf von auxWin[]:*/
  if (nWins > WINADMIN) 
  { printf ("\n Error in InitAuxWins!"); _getch(); return; }

  /*Quadratische Hilfsfenster:*/
  glutInitWindowSize(width, height);

  for (jj=0; jj<nWins; jj++)
  { glutInitWindowPosition (0, jj*(height+40)+20); //heuristisch ermittelt
    auxWin[jj] = glutCreateWindow("Results"); 
    glutDisplayFunc (callback); 
    DrawIn2D (width, height);
    glClearColor(1.0, 1.0, 1.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);

    /*Tastatur-Eingaben ueberall entgegennehmen:*/
    glutKeyboardFunc(Keyboard); 
  }
  /*Vorerst Hauptfenster aktiv:*/
  glutSetWindow(mainWin);
  return;
}
