/*FFTops.c (A. Christidis)*/

#include "LoadImg.h"
#include "DIPops.h"
#include "ShowAux.h"
#include "FFTops.h"

//#define SHIFT4CUT
#define CHK4BYTE(x)  (MIN(255,MAX(0,x)))

/*Merke: extern darf nicht initialisiert werden! (Deklaration vs. Definition)*/
extern   int    maskW, maskH;
extern   float *mask;

/***************************************************************************/
  void Fspec (DIPimg *srcImg, DIPimg *dstImg, int purpose)
/***************************************************************************/
/*Fourier spectrum of a monochrome image in several versions. Typically, 
  the source image <src> will be the original (to remain unchanged), the 
  destination image <dst> will be the work buffer to receive first the 
  b/w image, and then its Fourier spectrum.*/
{          int    jx=0, jy=0, idx=0, center=0;
  unsigned int    uipixel=0.f;
  unsigned char   *pixel=(unsigned char *)&uipixel; 
           float *ReImg=NULL, *ImImg=NULL, scale=0.f;
           float  Fmin=0.f, FMax=0.f;

  /*Pruefen, ob Bildmasse Potenzen von 2 sind:*/
  if (!checkSize (srcImg->Width) || !checkSize (srcImg->Height))
     { fprintf(stderr, "Dimensions must be powers of 2!\n" 
                       "(Press any key to continue)\n");  getch();  return; }

  /*Bild in monochrom konvertieren, weiter mit dstImg->Data:*/
  cv2grey (srcImg, dstImg);

  /*Speicher fuer Bild in float:*/
  ReImg = (float *) calloc(srcImg->Height*srcImg->Width, sizeof(float));
  ImImg = (float *) calloc(srcImg->Height*srcImg->Width, sizeof(float));

  /*Bild in float umspeichern und (zu Kontrollzwecken) aufrichten:*/
  for(jy=0; jy<srcImg->Height; jy++)
  { for (jx=0; jx<srcImg->Width; jx++) 
    { uipixel = dstImg->Data[jy*srcImg->Width+jx];
      ReImg[(srcImg->Height-jy-1)*srcImg->Width+jx] = (float) pixel[0]; 
    }
  }

#ifdef  SHIFT4CUT
  /*Bild auf zentriertes Spektrum vorbereiten (mit (-1)^(x+y) multiplizieren):*/
  if (center) centerSpec (ReImg, ReImg, srcImg->Width, srcImg->Height);
#else //SHIFT4CUT
#endif//SHIFT4CUT

  /*Bild-Transformation:*/
  fft2D (FWD, srcImg->Width, srcImg->Height, ReImg, ImImg);

  /*Betrag im Realteil ablegen, dort Extreme Fmin, FMax ermitteln:*/
  for(jy=0; jy<srcImg->Height; jy++)
  { for (jx=0; jx<srcImg->Width; jx++) 
    { 
#ifdef  MORE_FFT
#endif//MORE_FFT
    }
  }

  /*Spektrum-Versionen:*/ 
  switch (purpose)
  { case LIN: scale = 255.f / MAX (1.f, (FMax-Fmin)); 
              break;
#ifdef  MORE_FFT
#endif//MORE_FFT
  }

#ifdef  SHIFT4CUT
#else //SHIFT4CUT
  /*Ergebnisbild zentrieren:*/
  if (center) Center (ReImg, ReImg, srcImg->Width, srcImg->Height);
#endif//SHIFT4CUT

  /*Pixel-Darstellung (kopfunter):*/
  for(jy=0; jy<srcImg->Height; jy++)
  { for (jx=0; jx<srcImg->Width; jx++) 
    { idx = jy*srcImg->Width+jx;
      uipixel = (unsigned int) ReImg[idx];
      switch (purpose)
      { case LIN: uipixel = (unsigned char) (.5 + (ReImg[idx]-Fmin) * scale);
                  break;
#ifdef  MORE_FFT
#endif//MORE_FFT
      }
      pixel[2] = pixel[1] = pixel[0] = CHK4BYTE(uipixel);
      dstImg->Data[(srcImg->Height-jy-1)*srcImg->Width+jx] = uipixel;
    }
  }

  /*Freigabe des Speichers:*/
  free (ReImg);
  free (ImImg);
}

/***************************************************************************/
  void Ffilt (DIPimg *srcImg, DIPimg *dstImg, int purpose)
/***************************************************************************/
/*Filter image <srcImg> to destination image <dstImg> by convolving it with 
  a mask loaded by <loadMask>. <purpose> indicates the particular operation 
  to be executed. 
*/
{          int    jx=0, jy=0, idx=0, dummyi=0;
  unsigned int    uipixel=0;
  unsigned char  *pixel=(unsigned char *)&uipixel; 
           float *ReImg=NULL, *ImImg=NULL, *RePSF=NULL, *ImPSF=NULL, 
                  dummyf=0.f, dumRe=0.f, dumIm=0.f, eps=.00015;
           int    padW=0, padH=0;

  /*Pruefen, ob Bildmasse Potenzen von 2 sind:*/
  if (!checkSize (srcImg->Width) || !checkSize (srcImg->Height))
     { fprintf(stderr, "Dimensions must be powers of 2!\n" 
                       "(Press any key to continue)\n");  getch();  return; }

  /*Bild in monochrom konvertieren, weiter mit dstImg->Data:*/
  cv2grey (srcImg, dstImg);

  /*Gemeinsame Groesse Bild/PSF, Vorbereitung fuer Bild-Padding:*/
  padW = srcImg->Width;    padH = srcImg->Height;

  /*Speicher fuer Bild in float:*/
  ReImg = (float *) calloc(srcImg->Width * srcImg->Height, sizeof(float));
  ImImg = (float *) calloc(srcImg->Width * srcImg->Height, sizeof(float));

  /*Bild in float umspeichern und (zu Kontrollzwecken) aufrichten:*/
  for(jy=0; jy<srcImg->Height; jy++)
  { for (jx=0; jx<srcImg->Width; jx++) 
    { uipixel =  srcImg->Data[(srcImg->Height-jy-1)*srcImg->Width+jx];
      ReImg[jy*srcImg->Width + jx] = (float) pixel[0];
    } 
  } 

  /*Laden, Normieren der PSF, Masken-Check:*/
  loadMask (&mask, &maskH, &maskW);
  if (maskH > srcImg->Height || maskW > srcImg->Width) 
  { fprintf(stderr, "Mask cannot be greater than image!\n" 
                    "(Press any key to continue)\n");  getch();  return; }

  normMask (mask, maskH, maskW);

  /*Speicher fuer PSF:*/
  RePSF = mask;
  ImPSF = (float *) calloc(maskW * maskH, sizeof(float));
  if (!ImPSF) { fprintf(stderr, "ImPSF out of memory!\n"); _getch();  return; }

  /*PSF-Padding:*/
  padMat (&RePSF, maskW, maskH, &RePSF, padW, padH, FWD);
  padMat (&ImPSF, maskW, maskH, &ImPSF, padW, padH, FWD);


  /*Bild-Transformation:*/
  fft2D (FWD, padW, padH, ReImg, ImImg);

  /*PSF-Transformation:*/
  fft2D (FWD, padW, padH, RePSF, ImPSF);

  /*Faltung bzw. Filterung:*/
  switch (purpose)
  { case CONV: /*Faltung im Frequenzbereich: (a+bj)(c+dj)=(ac-bd) + j(ad+bc) */
               for (jy=0; jy<padH; jy++)
               { for (jx=0; jx<padW; jx++) 
                 { idx = jy*padW + jx;
#ifdef  MORE_FFT
#endif//MORE_FFT
                 }
               }
               break;

    case INVR: /*Inverse Filterung: 
                (a+bj)/(c+dj) = (ac+bd)/(c2+d2) + j(bc-ad)/(c2+d2) */
               for (jy=0; jy<padH; jy++)
               { for (jx=0; jx<padW; jx++) 
                 { idx= jy*padW + jx;
#ifdef  MORE_FFT
#endif//MORE_FFT
                 }
               }
               break;
  }

  /*Ergebnisbild im Frequenzbereich zentrieren:*/
#ifdef  SHIFT4CUT
  centerSpec (ReImg, ReImg, padW, padH);
  centerSpec (ImImg, ImImg, padW, padH);
#else //SHIFT4CUT
#endif//SHIFT4CUT

  /*Bild-Ruecktransformation:*/
  fft2D (INV, padW, padH, ReImg, ImImg);

  /*Alternativ: Ergebnisbild im Ortsbereich zentrieren:*/
#ifdef  SHIFT4CUT
#else //SHIFT4CUT
  Center (ReImg, ReImg, padW, padH);
#endif//SHIFT4CUT

  /*Pixel-Darstellung (kopfunter):*/
  for(jy=0; jy<srcImg->Height; jy++)
  { for (jx=0; jx<srcImg->Width; jx++) 
    { dummyi = (int) (ReImg[(jy)*srcImg->Width+ jx] + .5);
      uipixel = CHK4BYTE(dummyi);
      pixel[2] = pixel[1] = pixel[0] = CHK4BYTE(uipixel);
      dstImg->Data[(srcImg->Height-jy-1)*srcImg->Width+jx] = uipixel; 
    }
  }

  /*Freigabe des Speichers:*/
  free (ReImg);
  free (ImImg);
  free (RePSF);
  free (ImPSF);
}

/***************************************************************************/
  void centerSpec (float *src, float *dst, int w, int h)
/***************************************************************************/
/*Multiply elements of matrix or picture <src> by (-1)^(x+y), typically for 
  centering its visualized DFT spectrum. Source <src> and destination <dst> 
  pointers point to matrices of the same width <w> and height <h>; they may 
  point to identical data.*/
{ int jx=0, jy=0;

#define SIMPLE_MINDED
#ifdef  SIMPLE_MINDED
  /*Multiplikation mit pow((-1), (float)(jx+jy):*/
  for(jy=0; jy<h; jy++)
    for (jx=0; jx<w; jx++) 
      dst[jy*w+jx] =  src[jy*w+jx] * pow((-1.f), (float)(jx+jy));
#else //SIMPLE_MINDED
#ifdef  MORE_FFT
#endif//MORE_FFT
#endif//SIMPLE_MINDED

  return;
}

/***************************************************************************/
  void Center (float *src, float *dst, int w, int h)
/***************************************************************************/
/*Center source matrix <src> of width <w> and height <h> to destination 
  matrix <dst> by quartering it. 
  Pointers <src> and <dst> may point to identical data.*/
{ int   h2=0, w2=0, jx=0, jy=0;
  float help=0.f;

  if (w > 1) 
  { w2=w/2;
    for (jy=0; jy<h; jy++)
    { for (jx=0; jx<w2; jx++) 
      { help   = src[jy*w+jx];
        dst[jy*w+jx] = src[jy*w+jx+w2];
        dst[jy*w+jx+w2] = help;
      }
    }
  }
#ifdef  MORE_FFT
#endif//MORE_FFT
  return;
}

/***************************************************************************/
  void padMat (float **src, int srcW, int srcH, 
               float **dst, int dstW, int dstH, int mode)
/***************************************************************************/
/*"Pad" a source matrix <src> of size <srcW> x <srcH> to size <dstW> x <dstH> 
  and store it in the middle of the destination matrix <dst>, or, conversely, 
  "de-pad" it (cut out). Parameter <mode> is reserved for future development.
  Pointers <src> and <dst> may point to identical data. */
{
#ifdef  MORE_FFT
#endif//MORE_FFT

  /*Bei gleichen Dimensionen (Hoehe, Breite) von <dst> und <src> entbehrlich:*/
  if (srcW == dstW && srcH == dstH) return; 

#ifdef  MORE_FFT
#endif//MORE_FFT

  return;
}



/***************************************************************************/
  void fft2D (int Forward, int ImWid, int ImHig, float *ReImg, float *ImImg)
/***************************************************************************/
/* C realization of a two-dimensional, Discrete Fourier Transform.
   If Forward is 0 the procedure performs the inverse transform. */
{
   int    r, c, ImSize=MAX(ImWid, ImHig);
   float  *VecRe;
   float  *VecIm;

   VecRe = (float *) malloc (ImSize*sizeof(float));
   VecIm = (float *) malloc (ImSize*sizeof(float));

   for (r=0; r<ImHig; r++) {
      for (c=0; c<ImWid; c++) {
         VecRe[c] = ReImg[r*ImWid+c];
         VecIm[c] = ImImg[r*ImWid+c];
      }
      fft (Forward, ImWid, VecRe, VecIm);
      for (c=0; c<ImWid; c++) {
         ReImg[r*ImWid+c] = VecRe[c];
         ImImg[r*ImWid+c] = VecIm[c];
   }  }

   for (c=0; c<ImWid; c++) {
      for (r=0; r<ImHig; r++) {
         VecRe[r] = ReImg[r*ImWid+c];
         VecIm[r] = ImImg[r*ImWid+c];
      }
      fft (Forward, ImHig, VecRe, VecIm);
      for (r=0; r<ImHig; r++) {
         ReImg[r*ImWid+c] = VecRe[r];
         ImImg[r*ImWid+c] = VecIm[r];
   }  }
   free (VecRe);
   free (VecIm);
}

/***************************************************************************/
  void fft (int Forward, int Size, float *VecRe, float *VecIm)
/***************************************************************************/
/* C realization of the Fast Fourier Transform. 
   If Forward is 0 the procedure performs the inverse transform. */
{  const  double  PI = atan(1.)*4.;

   int    LenHalf, Stage, But, ButHalf, i,j,k, ip, pot2;
   float  ArcRe,ArcIm, dArcRe,dArcIm, ReBuf,ImBuf, ArcBuf;
   double Arc;
   pot2 = 0;
   while (Size != (1 << pot2))  pot2++;
   LenHalf = Size >> 1 ;
   j = 1;
   for (i=1; i<Size; i++) {
      if (i<j) {
         ReBuf = VecRe[j-1];
         ImBuf = VecIm[j-1];
         VecRe[j-1] = VecRe[i-1];
         VecIm[j-1] = VecIm[i-1];
         VecRe[i-1] = ReBuf;
         VecIm[i-1] = ImBuf;
      }
      k = LenHalf;
      while (k<j) {
         j -= k;  k = k >> 1;
      }
      j += k;
   }
   for (Stage=1; Stage<=pot2; Stage++) {
      But = 1 << Stage;
      ButHalf = But >> 1;
      ArcRe = (float)1;
      ArcIm = (float)0;
      Arc = (double) (PI/ButHalf);
      dArcRe = (float) cos(Arc);
      dArcIm = (float) sin(Arc);
      if (Forward)  dArcIm = -dArcIm;
      for (j=1; j<=ButHalf; j++) {
         i = j;
         while (i<=Size) {
            ip = i + ButHalf;
            ReBuf = VecRe[ip-1] * ArcRe - VecIm[ip-1] * ArcIm;
            ImBuf = VecRe[ip-1] * ArcIm + VecIm[ip-1] * ArcRe;
            VecRe[ip-1] = VecRe[i-1] - ReBuf;
            VecIm[ip-1] = VecIm[i-1] - ImBuf;
            VecRe[i-1]  = VecRe[i-1] + ReBuf;
            VecIm[i-1]  = VecIm[i-1] + ImBuf;
            i += But ;
         }
         ArcBuf = ArcRe;
         ArcRe = ArcRe  * dArcRe - ArcIm * dArcIm;
         ArcIm = ArcBuf * dArcIm + ArcIm * dArcRe;
   }  }
#define DIV_AT_INV
#ifdef  DIV_AT_INV
   if (!Forward) {
#else //DIV_AT_INV
   if (Forward) {
#endif//DIV_AT_INV
      for (j=1; j<=Size; j++) {
         VecRe[j-1] /= Size;
         VecIm[j-1] /= Size;
}  }  }

