/*                                 tga2rgb.c                               */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <gl/gl.h>

#include "tga2rgb.h"
#pragma warning(disable : 4996) // ("veraltet" nicht mehr melden!)


/* *********************************************************************** */
/* =============
   checkSize: Make sure it's a power of 2 (AC-Version).
   =============
*/
int checkSize (int x)
{
  int j=1;
  while (j < x) j <<= 1; 
  if (j == x) return (x);
  return (0);
}

/* *********************************************************************** */
/* =============
   returnError: Called when there is an error loading the .tga file.
   =============
*/
int returnError (FILE *filnam, int error)
{
    fclose (filnam);  return error;
}

/* *********************************************************************** */
/* =============
   getRGBA: Reads in RGBA data for a 32bit image. 
   =============
*/
unsigned int *getRGBA (FILE *filnam, int *imgWid, int *imgHig, int *format, 
                       int orient)
{
    unsigned char *rgba, temp;
    int            bread=0, i=0, pixel=0, size=(*imgWid)*(*imgHig);
    long int       linlen=0, filpos=0, bufpos=0;

    switch (*format) 
    {
      case GL_RGBA:   pixel = 4; break;
      case GL_RGB:    pixel = 3; break;
      case GL_ALPHA:  pixel = 1; break;
      default: returnError (filnam, TGA_BAD_BITS);
    }
    rgba = (unsigned char *) malloc (size * pixel); if (rgba == NULL) return 0;
    if (orient == BOTTOMUP)
      bread = fread (rgba, sizeof (unsigned char), size*pixel, filnam); 
    else // (orient == TOPDOWN)
    { linlen = (*imgWid)*pixel*sizeof (unsigned char);
      filpos = ftell (filnam) + size*pixel*sizeof (unsigned char); //End of Image
      for (i=(*imgHig); i > 0; i-- )
      { filpos -= linlen;
        fseek (filnam, filpos, SEEK_SET);	
        bread += fread (rgba+bufpos, sizeof(unsigned char),(*imgWid)*pixel,filnam); 
        bufpos += linlen;
      }
    }
    if (bread != size * pixel) { free (rgba); return 0; }

    /* TGA is stored in BGRA, make it RGBA */
    if ((*format)==GL_RGBA || (*format)==GL_RGB)
    { for (i = 0; i < size * pixel; i += pixel)
      { temp = rgba[i];  rgba[i] = rgba[i + 2];  rgba[i + 2] = temp; }
    }

    return (int *)rgba;
}

/* *********************************************************************** */
/* =============
   loadTGA: Loads up a targa file.  
            Supported types are 8,24 and 32 uncompressed images.  
            id is the texture ID to bind to.
   =============
*/
unsigned int *loadTGA (char *name, int *imgWid, int *imgHig, int *components,
                       int *format)
{
    unsigned char type[4];
    unsigned char info[7];
    unsigned int *imgDat = NULL, orient=BOTTOMUP;
    int bpp, size;
    FILE *filnam=NULL;

    if (!(filnam = fopen (name, "rbt"))) returnError (filnam, TGA_FILE_NOT_FOUND);

    fread (&type, sizeof (char), 3, filnam); // read in colormap info and image type, byte 0 ignored
    fseek (filnam, 12, SEEK_SET);			 // seek past the header and useless info
    fread (&info, sizeof (char), 6, filnam);
    if (type[1] != 0 || (type[2] != 2 && type[2] != 3))
        returnError (filnam, TGA_BAD_IMAGE_TYPE);

    *imgWid = info[0] + info[1] * 256; 
    *imgHig = info[2] + info[3] * 256;
    bpp =	info[4]; 
    if (info[5] & 0x20) orient=TOPDOWN;

    size = (*imgWid) * (*imgHig); 

    /* make sure dimension is a power of 2 */
    if (!checkSize (*imgWid) || !checkSize (*imgHig))
    { fprintf(stderr, "Dimensions should be power of 2!\n" 
                      "(Press any key to continue)\n"); 
    getchar(); }

    /* make sure we are loading a supported type */
    switch (bpp) 
    {
      case 32: (*format) = GL_RGBA;   *components=4; break;
      case 24: (*format) = GL_RGB;    *components=3; break;
      case  8: (*format) = GL_ALPHA;  *components=1; break;
      default: returnError (filnam, TGA_BAD_BITS);
    }
    imgDat = getRGBA (filnam, imgWid, imgHig, format, orient);

    /* no image data */
    if (imgDat == NULL) returnError (filnam, TGA_BAD_DATA);

    fclose (filnam);
    return imgDat;
}

